.ToastLink_toastLink__OLSeq{text-decoration:underline;font-size:var(--fidesui-font-size-md);padding-left:0;padding-right:0;color:var(--fidesui-minos)}.ToastLink_toastLink__OLSeq:focus-visible,.ToastLink_toastLink__OLSeq:hover{color:var(--fidesui-neutral-500)}