(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2992],{84983:function(n,e,i){(window.__NEXT_P=window.__NEXT_P||[]).push(["/settings/messaging-providers/new",function(){return i(51192)}])},51192:function(n,e,i){"use strict";i.r(e),i.d(e,{default:function(){return c}});var s=i(24246),t=i(96306),r=i(77213),u=i(90980),o=i(77830),f=i(10792);let a=()=>(0,s.jsxs)(r.Z,{title:"Create Messaging Configuration",children:[(0,s.jsx)(u.Z,{backPath:o.AD}),(0,s.jsx)(t.X6q,{fontSize:"2xl",fontWeight:"semibold",mb:6,children:"Configure your messaging provider"}),(0,s.jsx)(f.Z,{})]});var c=()=>(0,s.jsx)(a,{})}},function(n){n.O(0,[431,8765,9278,3662,9037,502,7158,2888,9774,179],function(){return n(n.s=84983)}),_N_E=n.O()}]);