(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2447],{64877:function(n,_,u){(window.__NEXT_P=window.__NEXT_P||[]).push(["/settings/about",function(){return u(71115)}])}},function(n){n.O(0,[1115,2888,9774,179],function(){return n(n.s=64877)}),_N_E=n.O()}]);