(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5988],{94896:function(n,u,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/settings/about/alpha",function(){return t(38292)}])},38292:function(n,u,t){"use strict";t.r(u);var e=t(24246),s=t(71115);u.default=()=>(0,e.jsx)(s.default,{showAlphaFeatures:!0})}},function(n){n.O(0,[1115,2888,9774,179],function(){return n(n.s=94896)}),_N_E=n.O()}]);