(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[165],{21865:function(e,t,a){(window.__NEXT_P=window.__NEXT_P||[]).push(["/data-catalog/[systemId]/projects",function(){return a(98791)}])},79609:function(e,t,a){"use strict";a.d(t,{N4:function(){return n},UG:function(){return o},xw:function(){return i}});var s=a(54427);let{useGetCatalogSystemsQuery:o,useGetCatalogProjectsQuery:i,useGetCatalogDatasetsQuery:n}=a(78780).u.injectEndpoints({endpoints:e=>({getCatalogSystems:e.query({query:e=>({method:"GET",url:"/plus/data-catalog/system",params:e}),providesTags:["Catalog Systems","System"]}),getCatalogProjects:e.query({query:e=>{let{...t}=e;return{method:"GET",url:"/plus/data-catalog/project",params:t}},providesTags:["Discovery Monitor Results"]}),getCatalogDatasets:e.query({query:e=>{let{...t}=e;return{method:"GET",url:"/plus/data-catalog/dataset",params:t}},providesTags:["Discovery Monitor Results"]})})});(0,s.oM)({name:"dataCatalog",initialState:{page:1,pageSize:50},reducers:{}})},98791:function(e,t,a){"use strict";a.r(t),a.d(t,{default:function(){return S}});var s=a(24246),o=a(86677),i=a(18225),n=a(77213),r=a(77830),l=a(58754),u=a(92222),d=a(59003),c=a(96306),g=a(27378),m=a(47935),p=a(72625),f=a(89200),x=a(5945),h=a(79609),j=a(99763),y=a(28976),C=a(8151);let w={items:[],total:0,page:1,size:50,pages:1},_=()=>(0,s.jsx)(c.gCW,{mt:6,p:10,spacing:4,borderRadius:"base",maxW:"70%","data-testid":"empty-state",alignSelf:"center",margin:"auto",children:(0,s.jsx)(c.gCW,{children:(0,s.jsx)(c.xvT,{fontSize:"md",fontWeight:"600",children:"No resources found"})})}),v=(0,u.Cl)();var R=e=>{let{systemKey:t,monitorConfigIds:a}=e,[i,n]=(0,g.useState)({}),{PAGE_SIZES:l,pageSize:c,setPageSize:R,onPreviousPageClick:b,isPreviousPageDisabled:S,onNextPageClick:P,isNextPageDisabled:E,startRange:k,endRange:D,pageIndex:N,setTotalPages:T}=(0,m.oi)(),{isFetching:M,isLoading:z,data:I}=(0,h.xw)({page:N,size:c,monitor_config_ids:a}),Z=(0,o.useRouter)(),{items:q,total:G,pages:V}=(0,g.useMemo)(()=>null!=I?I:w,[I]);(0,g.useEffect)(()=>{T(V)},[V,T]);let[W,X]=(0,g.useState)(void 0),U=(0,g.useMemo)(()=>[v.accessor(e=>e.name,{id:"name",cell:e=>(0,s.jsx)(f.Z,{resource:e.row.original}),header:"Project"}),v.display({id:"status",cell:e=>{let{row:t}=e;return(0,s.jsx)(x.Z,{status:(0,C.u)(t.original)})},header:"Status"}),v.accessor(e=>e.monitor_config_id,{id:"monitorConfigId",cell:e=>(0,s.jsx)(m.G3,{value:e.getValue()}),header:"Detected by"}),v.accessor(e=>e.description,{id:"description",cell:e=>(0,s.jsx)(m.G3,{value:e.getValue()}),header:"Description"}),v.accessor(e=>e.updated_at,{id:"lastUpdated",cell:e=>(0,s.jsx)(p.Cy,{time:e.getValue()}),header:"Updated",meta:{cellProps:{borderRight:"none"}}}),v.display({id:"actions",cell:e=>{let{row:t}=e;return(0,s.jsx)(y.Z,{onDetailClick:()=>X(t.original)})},size:25,meta:{disableRowClick:!0}})],[]),K=(0,d.b7)({getCoreRowModel:(0,u.sC)(),getGroupedRowModel:(0,u.qe)(),getExpandedRowModel:(0,u.rV)(),manualPagination:!0,columnResizeMode:"onChange",columns:U,data:q,getRowId:e=>e.urn,onRowSelectionChange:n,state:{rowSelection:i}});return z||M?(0,s.jsx)(m.I4,{rowHeight:36,numRows:36}):(0,s.jsxs)(s.Fragment,{children:[(0,s.jsx)(m.ZK,{tableInstance:K,emptyTableNotice:(0,s.jsx)(_,{}),onRowClick:e=>Z.push("".concat(r.mX,"/").concat(t,"/projects/").concat(e.urn))}),(0,s.jsx)(m.s8,{totalRows:G||0,pageSizes:l,setPageSize:R,onPreviousPageClick:b,isPreviousPageDisabled:S,onNextPageClick:P,isNextPageDisabled:E,startRange:k,endRange:D}),(0,s.jsx)(j.Z,{resource:W,onClose:()=>X(void 0)})]})},b=a(1315),S=()=>{var e;let{query:t}=(0,o.useRouter)(),a=t.systemId,u=t.monitor_config_ids,{data:d,isLoading:c}=(0,b.rn)(a);return c?(0,s.jsx)(i.Z,{}):(0,s.jsxs)(n.Z,{title:"Data catalog",children:[(0,s.jsx)(l.Z,{heading:"Data catalog",breadcrumbItems:[{title:"All systems",href:r.mX},{title:null!==(e=null==d?void 0:d.name)&&void 0!==e?e:null==d?void 0:d.fides_key}]}),(0,s.jsx)(R,{systemKey:a,monitorConfigIds:u})]})}}},function(e){e.O(0,[8765,9278,5163,409,6882,2888,9774,179],function(){return e(e.s=21865)}),_N_E=e.O()}]);