<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <title>Fides Test message</title>
  </head>
  <body>
    <main>
      <p>This is a test message from Fides.</p>
    </main>
  </body>
</html>
