<!DOCTYPE html>
<html lang="en">
   <head>
      <meta charset="UTF-8">
      <title>Erasure requests</title>
   </head>
   <body>
      <main>
         <p>
            Following users of {{controller}} have requested to delete their data. You are notified of the changes because {{third_party_vendor_name}} has been identified as a third-party processor to {{controller}} that processes user information.
         </p>
         <p>Please find below the list of users:</p>
         <ul>
            {% for identity in identities -%}
            <li>{{identity}}</li>
            {%- endfor %}
         </ul>
         <p>You are legally obligated to honor the user's data deletion request.</p>
      </main>
   </body>
</html>