"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2121],{22121:function(n,e,t){async function s(){{let{worker:n}=await Promise.all([t.e(5698),t.e(7202)]).then(t.bind(t,57202));await n.start({onUnhandledRequest:"bypass"})}}t.r(e),t.d(e,{initMocks:function(){return s}})}}]);