(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[9851],{66393:function(e,n,i){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent",function(){return i(78220)}])},77213:function(e,n,i){"use strict";i.d(n,{Z:function(){return f}});var t=i(24246),s=i(96306),r=i(88038),o=i.n(r),a=i(86677);i(27378);var g=i(25980),u=i(90867),c=i(42478),d=i(77830),l=()=>{let e=(0,a.useRouter)();return(0,t.jsx)(s.xuv,{bg:"gray.50",border:"1px solid",borderColor:"blue.400",borderRadius:"md",justifyContent:"space-between",p:5,mb:5,mt:5,children:(0,t.jsxs)(s.xuv,{children:[(0,t.jsxs)(s.Kqy,{direction:{base:"column",sm:"row"},justifyContent:"space-between",children:[(0,t.jsx)(s.xvT,{fontWeight:"semibold",children:"Configure your storage and messaging provider"}),(0,t.jsx)(s.wpx,{onClick:()=>{e.push(d.AD)},children:"Configure"})]}),(0,t.jsxs)(s.xvT,{children:["Before Fides can process your privacy requests we need two simple steps to configure your storage and email client."," "]})]})})},f=e=>{let{children:n,title:i,padded:r=!0,mainProps:d}=e,f=(0,g.hz)(),y=(0,a.useRouter)(),m="/privacy-requests"===y.pathname||"/datastore-connection"===y.pathname,C=!(f.flags.messagingConfiguration&&m),{data:p}=(0,c.JE)(void 0,{skip:C}),{data:v}=(0,u.PW)(void 0,{skip:C}),h=f.flags.messagingConfiguration&&(!p||!v)&&m;return(0,t.jsxs)(s.kCb,{"data-testid":i,direction:"column",h:"100vh",children:[(0,t.jsxs)(o(),{children:[(0,t.jsxs)("title",{children:["Fides Admin UI - ",i]}),(0,t.jsx)("meta",{name:"description",content:"Privacy Engineering Platform"}),(0,t.jsx)("link",{rel:"icon",href:"/favicon.ico"})]}),(0,t.jsxs)(s.kCb,{as:"main",direction:"column",py:r?6:0,px:r?10:0,h:r?"calc(100% - 48px)":"full",flex:1,minWidth:0,overflow:"auto",...d,children:[h?(0,t.jsx)(l,{}):null,n]})]})}},42478:function(e,n,i){"use strict";i.d(n,{FU:function(){return u},JE:function(){return s},Ki:function(){return d},SU:function(){return l},W:function(){return f},h9:function(){return r},jc:function(){return t},qt:function(){return g},sn:function(){return c}});let{useGetEmailInviteStatusQuery:t,useGetActiveMessagingProviderQuery:s,useCreateMessagingConfigurationMutation:r,useCreateMessagingConfigurationSecretsMutation:o,useGetMessagingConfigurationDetailsQuery:a,useGetMessagingConfigurationsQuery:g,useGetMessagingConfigurationByKeyQuery:u,useUpdateMessagingConfigurationByKeyMutation:c,useUpdateMessagingConfigurationSecretsByKeyMutation:d,useCreateTestConnectionMessageMutation:l,useDeleteMessagingConfigurationByKeyMutation:f}=i(78780).u.injectEndpoints({endpoints:e=>({getEmailInviteStatus:e.query({query:()=>({url:"/messaging/email-invite/status"}),providesTags:()=>["Email Invite Status"]}),createMessagingConfiguration:e.mutation({query:e=>({url:"messaging/config",method:"POST",body:e}),invalidatesTags:["Messaging Config","Configuration Settings"]}),getActiveMessagingProvider:e.query({queryFn:async(e,n,i,t)=>{let s=await t({url:"messaging/default/active"});return s.error&&404===s.error.status?{data:null}:s},providesTags:["Messaging Config"]}),createMessagingConfigurationSecrets:e.mutation({query:e=>({url:"messaging/default/".concat(e.service_type,"/secret"),method:"PUT",body:e.details}),invalidatesTags:["Messaging Config","Configuration Settings"]}),getMessagingConfigurationDetails:e.query({query:e=>({url:"messaging/default/".concat(e.type)}),providesTags:["Messaging Config"]}),getMessagingConfigurations:e.query({query:()=>({url:"messaging/config"}),providesTags:["Messaging Config"]}),getMessagingConfigurationByKey:e.query({query:e=>({url:"messaging/config/".concat(e.key)}),providesTags:["Messaging Config"]}),createTestConnectionMessage:e.mutation({query:e=>({url:"messaging/test/".concat(e.service_type),method:"POST",body:e.details}),invalidatesTags:["Messaging Config"]}),updateMessagingConfigurationByKey:e.mutation({query:e=>({url:"messaging/config/".concat(e.key),method:"PATCH",body:e.config}),invalidatesTags:["Messaging Config","Configuration Settings"]}),updateMessagingConfigurationSecretsByKey:e.mutation({query:e=>({url:"messaging/config/".concat(e.key,"/secret"),method:"PUT",body:e.secrets}),invalidatesTags:["Messaging Config","Configuration Settings"]}),deleteMessagingConfigurationByKey:e.mutation({query:e=>({url:"messaging/config/".concat(e.key),method:"DELETE"}),invalidatesTags:["Messaging Config","Configuration Settings"]})})})},78220:function(e,n,i){"use strict";i.r(n);var t=i(24246),s=i(96306),r=i(86677),o=i(27378),a=i(77213),g=i(77830);n.default=()=>{let e=(0,r.useRouter)();return(0,o.useEffect)(()=>{e.push(g.IT)},[e]),(0,t.jsx)(a.Z,{title:"Consent",children:(0,t.jsx)(s.M5Y,{children:(0,t.jsx)(s.$jN,{})})})}}},function(e){e.O(0,[2888,9774,179],function(){return e(e.s=66393)}),_N_E=e.O()}]);