(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[1285],{63641:function(e,n,i){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent/properties",function(){return i(63063)}])},63063:function(e,n,i){"use strict";i.r(n);var r=i(24246),s=i(96306),t=i(77213),o=i(58754),a=i(25783);n.default=()=>(0,r.jsxs)(t.Z,{title:"Properties",children:[(0,r.jsx)(o.Z,{heading:"Properties",children:(0,r.jsx)(s.xvT,{fontSize:"sm",width:{base:"100%",lg:"60%"},children:"Review and manage your properties below. Properties are the locations you have configured for consent management such as a website or mobile app."})}),(0,r.jsx)(a.V,{})]})}},function(e){e.O(0,[2040,5783,2888,9774,179],function(){return e(e.s=63641)}),_N_E=e.O()}]);