(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5970],{6920:function(e,l,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/poc/form-experiments/AntForm",function(){return n(33396)}])},95813:function(e,l,n){"use strict";n.d(l,{S:function(){return t}});let t={input:"default text",radio:"2",checkbox:!0,switch:!0,select:"2",multiselect:["2"],tags:["1"],date:"",number:12}},33396:function(e,l,n){"use strict";n.r(l),n.d(l,{AntFormPOC:function(){return u}});var t=n(24246),a=n(96306),s=n(3124),r=n.n(s),c=n(27378),i=n(95813);let{Title:o}=a.AntTypography,u=()=>{let[e]=a.PPS.useForm(),[l,n]=(0,c.useState)(i.S);return(0,t.jsxs)(a.bue,{children:[(0,t.jsxs)(a.JGx,{span:12,children:[(0,t.jsx)(o,{level:2,children:"Pure Ant"}),(0,t.jsxs)("div",{className:"mb-4",children:[(0,t.jsx)(a.j8w,{color:"green",children:"funcitonal"}),(0,t.jsx)(a.j8w,{color:"green",children:"low effort"})]}),(0,t.jsxs)(a.PPS,{initialValues:{...i.S},form:e,layout:"vertical",onValuesChange:(e,l)=>{n(l)},children:[(0,t.jsx)(a.PPS.Item,{label:"Input",name:"input",children:(0,t.jsx)(a.uFc,{})}),(0,t.jsx)(a.PPS.Item,{name:"radio",children:(0,t.jsxs)(a.y02.Group,{name:"radio",children:[(0,t.jsx)(a.y02,{value:"1",children:"Radio 1"}),(0,t.jsx)(a.y02,{value:"2",children:"Radio 2"})]})}),(0,t.jsx)(a.PPS.Item,{name:"checkbox",valuePropName:"checked",children:(0,t.jsx)(a.E_O,{children:"Checkbox"})}),(0,t.jsx)(a.PPS.Item,{label:"Switch",name:"switch",valuePropName:"checked",children:(0,t.jsx)(a.rAg,{})}),(0,t.jsx)(a.PPS.Item,{label:"Select",name:"select",children:(0,t.jsx)(a.WPr,{allowClear:!0,className:"w-full",options:[{value:"1",label:"Option 1"},{value:"2",label:"Option 2"}]})}),(0,t.jsx)(a.PPS.Item,{label:"Multiselect",name:"multiselect",children:(0,t.jsx)(a.WPr,{mode:"multiple",allowClear:!0,className:"w-full",options:[{value:"1",label:"Option 1"},{value:"2",label:"Option 2"}]})}),(0,t.jsx)(a.PPS.Item,{label:"Tags",name:"tags",children:(0,t.jsx)(a.WPr,{mode:"tags",allowClear:!0,className:"w-full",options:[{value:"1",label:"Option 1"},{value:"2",label:"Option 2"}]})}),(0,t.jsx)(a.PPS.Item,{label:"Date",name:"date",children:(0,t.jsx)(a.cv2,{})}),(0,t.jsx)(a.PPS.Item,{label:"Number",name:"number",children:(0,t.jsx)(a.W7D,{})})]})]}),(0,t.jsxs)(a.JGx,{span:8,offset:4,children:[(0,t.jsx)(o,{level:4,children:"Controlled Values"}),(0,t.jsx)(a.V4A,{style:{backgroundColor:r().FIDESUI_MINOS,color:r().FIDESUI_CORINTH},children:(0,t.jsx)("pre",{children:(0,t.jsxs)("code",{children:["{\n",Object.keys(l).map(e=>"  ".concat(e.toString(),": ").concat(l[e],"\n")),"}"]})})})]})]})};l.default=u}},function(e){e.O(0,[2888,9774,179],function(){return e(e.s=6920)}),_N_E=e.O()}]);